// These routines extend an alignment in a given direction (forward or
// reverse) from given start points in two sequences.

// Forward alignments begin at the given start points, whereas reverse
// alignments begin one-before the given start points.

// To use: first call "align", which calculates the alignment but only
// returns its score.  To get the actual alignment, call
// "getNextChunk" to get each gapless chunk.

// The sequences had better end with sentinels: the score for matching
// a sentinel with anything should be -INF.

// The gap parameters correspond to "generalized affine gap costs"
// (Proteins 1998 32(1):88-96).
// gapExistenceCost = a
// gapExtensionCost = b
// gapUnalignedCost = c
// When c >= a + 2b, it reduces to standard affine gap costs:
// gap cost = gapExistenceCost + gapExtensionCost * (gap length).

// The algorithm proceeds antidiagonal-by-antidiagonal, similarly to
// Section 2 in J Comput Biol. 2000 7(1-2):203-14.  It does not allow
// the score to drop by more than maxScoreDrop below the highest score
// in any previous antidiagonal.

// The parameter maxMatchScore should be the highest possible score
// for matching 2 letters.  This parameter is not actually necessary,
// but it provides some optimization opportunities.  If you give it a
// too-high value, the results will not change, but the run time may
// increase.

#ifndef GAPPED_XDROP_ALIGNER_HH
#define GAPPED_XDROP_ALIGNER_HH

//#include "ScoreMatrixRow.hh"

#include <cstddef>  // size_t
#include <vector>
#include <cassert>
#include <climits>

//typedef unsigned char uchar;

// The row size must be fixed to some value.  It is fixed to 64
// because: this is big enough for all amino acids, including
// ambiguous ones, in upper and lower case, and using a power-of-2
// might be fast.
enum { scoreMatrixRowSize = 64 };

typedef int ScoreMatrixRow[scoreMatrixRowSize];

// An "infinite" score.  Delimiters at the ends of sequences get a
// score of -INF.  We want it high enough to terminate alignments
// immediately, but not so high that it causes overflow errors.
enum { INF = INT_MAX / 3 };

class GappedXdropAligner {
 public:
  GappedXdropAligner(int gapExistCost=2,int gapExtendCost=3, int mmScore=4) {
    gapExistenceCost=gapExistCost;
    gapExtensionCost=gapExtendCost;
    maxMatchScore=mmScore;
    gapUnalignedCost = gapExistenceCost + (gapExtensionCost<<1);
    //isAffine = gapUnalignedCost >= gapExistenceCost + (gapExtensionCost<<1);
    isAffine = true; //for now
    }
  int align(const char *seq1_base,
      int seq1_from,
      int seq1_to,
      const char *seq2_base,
      int seq2_from,
      int seq2_to,
      bool isForward=true,
      //const ScoreMatrixRow *scorer,
      int maxScoreDrop=16);


  // Call this repeatedly to get each gapless chunk of the alignment.
  // The chunks are returned in far-to-near order.  The chunk's end
  // coordinates in each sequence (relative to the start of extension)
  // and length are returned in the first 3 parameters.  If there are
  // no more chunks, the 3 parameters are unchanged and "false" is
  // returned.
  bool getNextChunk(int &end1,
                    int &end2,
                    int &length);

  // The next 3 functions are for use by Centroid.  If the Centroid
  // code gets updated, it might make sense to change these functions too.

  // The number of antidiagonals, including dummy ones at the beginning.
  std::size_t numAntidiagonals() const
  { return seq1starts.size(); }

  std::size_t seq1start(std::size_t antidiagonal) const
  { return seq1starts[antidiagonal]; }

  std::size_t numCellsAndPads(std::size_t antidiagonal) const
  { return scoreEnds[antidiagonal + 1] - scoreEnds[antidiagonal]; }

  // The index in the score vectors, of the previous "horizontal" cell.
  int hori(std::size_t antidiagonal, std::size_t seq1coordinate) const
  { return (int)(scoreBase(antidiagonal - 1) + seq1coordinate); }

  // The index in the score vectors, of the previous "vertical" cell.
  int vert(std::size_t antidiagonal, std::size_t seq1coordinate) const
  { return (int)(scoreBase(antidiagonal - 1) + seq1coordinate + 1); }

  // The index in the score vectors, of the previous "diagonal" cell.
  int diag(std::size_t antidiagonal, std::size_t seq1coordinate) const
  { return (int)(scoreBase(antidiagonal - 2) + seq1coordinate); }

 private:
  std::vector<int> xScores;  // best score ending with aligned letters
  std::vector<int> yScores;  // best score ending with insertion in seq1
  std::vector<int> zScores;  // best score ending with insertion in seq2

  std::vector<std::size_t> seq1starts;  // seq1 start pos for each antidiagonal
  std::vector<std::size_t> scoreEnds;  // score end pos for each antidiagonal

  // Our position during the trace-back:
  std::size_t bestAntidiagonal;
  std::size_t bestSeq1position;

  int gapExistenceCost;
  int gapExtensionCost;
  int gapUnalignedCost;
  int maxMatchScore;
  bool isAffine;

  std::size_t scoreBase(std::size_t antidiagonal) const
  { return scoreEnds[antidiagonal] - seq1starts[antidiagonal]; }

  void init();

  void initAntidiagonal(std::size_t seq1beg, std::size_t scoreEnd,
                        std::size_t numCells);

  void updateBest(int &bestScore, int score, std::size_t antidiagonal,
                  const int *x0, const int *x0ori);

};


#endif
