#include "GBase.h"
#include "GArgs.h"
#include "GStr.h"
#include "GList.hh"
//#include "GXDropAlign.h"

#define USAGE "Usage:\n\
   seqalign -i <2line_file> | -a <seq1> -b <seq2>\n"

bool debug=true;

struct SEditScript{
    uint32 *op;               // array of edit operations
    uint32 size, num;         // size of allocation, number in use
    uint32 last;              // most recent operation added
};
/*
SEditScript *edit_script_free(SEditScript *es);
SEditScript *edit_script_new(void);
SEditScript *edit_script_append(SEditScript *es, SEditScript *et);
*/
enum {
    EDIT_OP_MASK = 0x3,
    EDIT_OP_ERR  = 0x0,
    EDIT_OP_INS  = 0x1,
    EDIT_OP_DEL  = 0x2,
    EDIT_OP_REP  = 0x3
};

//enum {         /* half of the (fixed) match score */
//    ERROR_FRACTION=2,  /* 1/this */
//    MAX_SPACE=1000000,
//    sC = 0, sI = 1, sD = 2, LARGE=100000000
//};

#define ERROR_FRACTION 2  /* 1/this */
#define MAX_SPACE 1000000
#define LARGE 100000000
int sC = 0, sI = 1, sD = 2;

uint32 edit_val_get(uint32 op) {
    return op >> 2;
}
uint32 edit_opc_get(uint32 op) {
    return op & EDIT_OP_MASK;
}

uint32 edit_op_cons(uint32 op, uint32 val) {
    return (val << 2) | (op & EDIT_OP_MASK);
}

uint32 edit_op_inc(uint32 op, uint32 n) {
    return edit_op_cons(edit_opc_get(op), edit_val_get(op) + n);
}

uint32 edit_op_inc_last(SEditScript *es, uint32 n) {
    uint32 *last;
    //ASSERT (es->num > 0);
    last = &(es->op[es->num-1]);
    *last = edit_op_inc(*last, n);
    return *last;
}

int edit_script_ready(SEditScript *es, uint32 n) {
    //uint32 *p;
    uint32 m = n + n/2;
    if (es->size <= n) {
        GREALLOC(es->op, m*sizeof(uint32));
        //es->op = p;
        es->size = m;
        }
    return 1;
}

int edit_script_readyplus(SEditScript *es, uint32 n) {
    if (es->size - es->num <= n)
        return edit_script_ready(es, n + es->num);
    return 1;
}

int edit_script_put(SEditScript *es, uint32 op, uint32 n) {
    if (!edit_script_readyplus(es, 2))
        return 0;
    es->last = op;
    //ASSERT(op != 0);
    es->op[es->num] = edit_op_cons(op, n);
    es->num += 1;
    es->op[es->num] = 0; /* sentinal */

    return 1;
}

SEditScript *edit_script_init(SEditScript *es) {
  es->op = 0;
  es->size = es->num = 0;
  es->last = 0;
  edit_script_ready(es, 8);
  return es;
}

uint32 *edit_script_first(SEditScript *es) {
    return es->num > 0 ? &es->op[0] : 0;
}

uint32 *edit_script_next(SEditScript *es, uint32 *op) {
    //  assumes flat address space !
    if (&es->op[0] <= op && op < &es->op[es->num-1])
        return op+1;
    else
        return 0;
}

static int edit_script_more(SEditScript *data, uint32 op, uint32 k) {
    if (op == EDIT_OP_ERR) {
        GError("edit_script_more: bad opcode %d:%d", op, k);
        return -1;
        }
    
    if (edit_opc_get(data->last) == op)
        edit_op_inc_last(data, k);
    else
        edit_script_put(data, op, k);

    return 0;
}

/* External */
SEditScript *edit_script_append(SEditScript *es, SEditScript *et) {
    uint32 *op;
    for (op = edit_script_first(et); op; op = edit_script_next(et, op))
        edit_script_more(es, edit_opc_get(*op), edit_val_get(*op));
    return es;
}

SEditScript *edit_script_new(void) {
    SEditScript *es = NULL;
    GCALLOC(es, sizeof(*es));
    return edit_script_init(es);
}

SEditScript *edit_script_free(SEditScript *es) {
    if (es) {
        if (es->op) {
            //MemFree(es->op);
            GFREE(es->op);
            }
        //MemSet(es, 0, sizeof(*es));
        //MemFree(es);
        GFREE(es);
        }
    return 0;
}

int edit_script_del(SEditScript *data, uint32 k) {
    return edit_script_more(data, EDIT_OP_DEL, k);
}

int edit_script_ins(SEditScript *data, uint32 k) {
    return edit_script_more(data, EDIT_OP_INS, k);
}
int edit_script_rep(SEditScript *data, uint32 k) {
    return edit_script_more(data, EDIT_OP_REP, k);
}

SEditScript *edit_script_reverse_inplace(SEditScript *es) {
    uint32 i;
    const uint32 num = es->num;
    const uint32 mid = num/2;
    const uint32 end = num-1;
    
    for (i = 0; i < mid; ++i) {
        const uint32 t = es->op[i];
        es->op[i] = es->op[end-i];
        es->op[end-i] = t;
    }
    return es;
}

/* ----- pool allocator ----- */
struct ThreeVal {
    int I, C, D;
};

struct MBSpace {
    ThreeVal* space_array;
    int used, size;
    MBSpace *next;
};

MBSpace* new_mb_space() {
    MBSpace* p;
    int amount;
    
    GMALLOC(p, sizeof(MBSpace));
    amount = MAX_SPACE;
    GCALLOC(p->space_array, sizeof(ThreeVal)*amount);
    if (p->space_array == NULL) {
       GFREE(p);
       return NULL;
    }
    p->used = 0; 
    p->size = amount;
    p->next = NULL;

    return p;
}

void refresh_mb_space(MBSpace* sp)
{
   while (sp) {
      sp->used = 0;
      sp = sp->next;
   }
}

void free_mb_space(MBSpace* sp)
{
   MBSpace* next_sp;
   while (sp) {
      next_sp = sp->next;
      //sp->space_array = MemFree(sp->space_array);
      GFREE(sp->space_array);
      //sp = MemFree(sp);
      GFREE(sp);
      sp = next_sp;
   }
}

ThreeVal* get_mb_space(MBSpace* S, int amount)
{
    ThreeVal* s;
    if (amount < 0) 
        return NULL;  
    
    while (S->used+amount > S->size) {
       if (S->next == NULL)
          if ((S->next = new_mb_space()) == NULL) {
            GError("Cannot get new space for greedy extension!\n");
            return NULL;
            }
       S = S->next;
    }
    s = S->space_array+S->used;
    S->used += amount;
    return s;
}

struct GreedyAlignMem {
  int** flast_d;
  int* max_row_free;
  int* uplow_free;
  MBSpace* space;
  int max_d;
  GreedyAlignMem(int max_len, int reward, int penalty, int Xdrop) {
    flast_d=NULL;
    max_row_free=NULL;
    max_d=0;
    uplow_free=NULL;
    space=NULL;
    max_d = (int)(max_len / ERROR_FRACTION + 1);
    int d_diff = (Xdrop+reward/2)/(penalty+reward) + 1;
    //flast_d = (int**) malloc((max_d + 2) * sizeof(int));
    GCALLOC(flast_d, (max_d+2)* sizeof(int));
    if (flast_d == NULL) {
       GError("Error: cannot allocate GreedyAlignMem structure!\n");
       }
    flast_d[0] = (int*)malloc((max_d + max_d + 6) * sizeof(int) * 2);
    if (flast_d[0] == NULL) {
        GError("Error: failed to allocate flast_d[0] in GreedyAlignMem!\n");
        }
    flast_d[1] = flast_d[0] + max_d + max_d + 6;
    //for (int i=2;i<max_d+2;i++) {
    //  flast_d[i]=NULL;
    //  }
    //search->abmp->flast_d_affine = NULL;
    uplow_free = NULL;
    max_row_free = (int*) malloc(sizeof(int) * (max_d + 1 + d_diff));
    space = new_mb_space();
    }
  ~GreedyAlignMem() {
    free_mb_space(space);
    GFREE(max_row_free);
    //GFREE(space);
    GFREE(flast_d[0]);
    //for (int i=0;i<max_d+2;i++) {
    //   GFREE(flast_d[i]);
    //   }
    
    GFREE(flast_d);
    }
};


int get_last(int **flast_d, int d, int diag, int *row1) {
  if (flast_d[d-1][diag-1] > GMAX(flast_d[d-1][diag], flast_d[d-1][diag+1])) {
      *row1 = flast_d[d-1][diag-1];
      return diag-1;
      } 
  if (flast_d[d-1][diag] > flast_d[d-1][diag+1]) {
      *row1 = flast_d[d-1][diag];
      return diag;
      }
  *row1 = flast_d[d-1][diag+1];
  return diag+1;
}


int MegaBlastGreedyAlign(const char* s1, int len1,
        const char* s2, int len2,
        bool reverse, int xdrop_threshold, 
        int match_cost, int mismatch_cost,
        int* e1, int* e2, 
        GreedyAlignMem* abmp, SEditScript *S)
                          // Uint1 rem) 
{
  int col,      /* column number */
      d,        /* current distance */
      k,        /* current diagonal */
      flower, fupper,            /* boundaries for searching diagonals */
      row,            /* row number */
      MAX_D,       /* maximum cost */
      ORIGIN,
      return_val = 0;
  int** flast_d = abmp->flast_d; /* rows containing the last d */
  int* max_row;    /* reached for cost d=0, ... len1.  */

  int X_pen = xdrop_threshold;
  int M_half = match_cost/2;
  int Op_cost = mismatch_cost + M_half*2;
  int D_diff = (X_pen+M_half) / Op_cost + 1;
  int x, cur_max, b_diag = 0, best_diag = INT_MAX/2;
  int* max_row_free = abmp->max_row_free;
  char nlower = 0, nupper = 0;
  MBSpace* space = abmp->space;
  int max_len = len2;

  MAX_D = (int) (len1/ERROR_FRACTION + 1);
  ORIGIN = MAX_D + 2;
  GMessage("D_diff=%d, MAX_D=%d\n", D_diff, MAX_D);

  *e1 = *e2 = 0;
   //find first mismatch
  if (reverse) {
     /*
     if (!(rem & 4)) {
        for (row = 0; row < len2 && row < len1 && 
                (s2[len2-1-row] ==
                 READDB_UNPACK_BASE_N(s1[(len1-1-row)/4], 
                                      3-(len1-1-row)%4)); 
             row++)
           ; //empty
     } else {
     */
        for (row = 0; row < len2 && row < len1 && (s2[len2-1-row] == s1[len1-1-row]); row++)
           ; //empty
     //}
    } else {
     /*
     if (!(rem & 4)) {
        for (row = 0; row < len2 && row < len1 && 
                (s2[row] == 
                 READDB_UNPACK_BASE_N(s1[(row+rem)/4], 
                                      3-(row+rem)%4)); 
             row++)
           ; //empty
     } else { */
        for (row = 0; row < len2 && row < len1 && (s2[row] == s1[row]); row++)
            ; //empty
     //}
  }
  *e1 = row;
  *e2 = row;
  if (row == len1) {
      if (S != NULL)
          edit_script_rep(S, row);
      // hit last row; stop search
      return 0;
      }
  if (S==NULL) 
     space = 0;
  else if (!space)
     abmp->space = space = new_mb_space();
  else 
     refresh_mb_space(space);

  max_row = max_row_free + D_diff;
  for (k = 0; k < D_diff; k++)
    max_row_free[k] = 0;

  flast_d[0][ORIGIN] = row;
  max_row[0] = (row + row)*M_half;

  flower = ORIGIN - 1;
  fupper = ORIGIN + 1;

  d = 1;
  while (d <= MAX_D) {
      int fl0, fu0;
      flast_d[d - 1][flower - 1] = flast_d[d - 1][flower] = -1;
      flast_d[d - 1][fupper] = flast_d[d - 1][fupper + 1] = -1;
      x = max_row[d - D_diff] + Op_cost * d - X_pen;
      x = (int)ceil(x/M_half);
      cur_max = 0;
      fl0 = flower;
      fu0 = fupper;
      for (k = fl0; k <= fu0; k++) {
          row = GMAX(flast_d[d - 1][k + 1], flast_d[d - 1][k]) + 1;
          row = GMAX(row, flast_d[d - 1][k - 1]);
          col = row + k - ORIGIN;
          if (row + col >= x)
              fupper = k;
            else {
              if (k == flower)
                  flower++;
              else
                  flast_d[d][k] = -1;
              continue;
              }
          if (row > max_len || row < 0) {
                flower = k+1; nlower = 1;
               } else {
               /* Slide down the diagonal. Don't do this if reached 
                  the end point, which has value 0x0f */
                if (reverse) {
                  if (s2[len2-row] != 0x0f) {
                     /*if (!(rem & 4))
                     while (row < len2 && col < len1 && s2[len2-1-row] == 
                            READDB_UNPACK_BASE_N(s1[(len1-1-col)/4],
                                                 3-(len1-1-col)%4)) {
                        ++row;
                        ++col;
                     }
                     else */
                      while (row < len2 && col < len1 && s2[len2-1-row] == s1[len1-1-col]) {
                        ++row;
                        ++col;
                        }
                  } else {
                     max_len = row;
                     flower = k+1; nlower = 1;
                     }
                } else if (s2[row-1] != 0x0f) {
                  /*if (!(rem & 4)) {
                     while (row < len2 && col < len1 && s2[row] == 
                            READDB_UNPACK_BASE_N(s1[(col+rem)/4],
                                                 3-(col+rem)%4)) {
                        ++row;
                        ++col;
                     }
                  } else { */
                     while (row < len2 && col < len1 && s2[row] == s1[col]) {
                        ++row;
                        ++col;
                        }
                  // }
                } else {
                  max_len = row;
                  flower = k+1; nlower = 1;
                  }
                }
          flast_d[d][k] = row;
          if (row + col > cur_max) {
            cur_max = row + col;
            b_diag = k;
            }
          if (row == len2) {
            flower = k+1; nlower = 1;
            }
          if (col == len1) {
            fupper = k-1; nupper = 1;
            }
          } //for k
      k = cur_max*M_half - d * Op_cost;
      if (max_row[d - 1] < k) {
          max_row[d] = k;
          return_val = d;
          best_diag = b_diag;
          *e2 = flast_d[d][b_diag];
          *e1 = (*e2)+b_diag-ORIGIN;
          } else {
          max_row[d] = max_row[d - 1];
          }
      if (flower > fupper)
          break;
      d++;
      if (!nlower) flower--; 
      if (!nupper) fupper++;
      if (S==NULL)
         flast_d[d] = flast_d[d - 2];
        else {
          // space array consists of ThreeVal structures which are 
          // 3 times larger than int, so divide requested amount by 3
         flast_d[d] = (int*) get_mb_space(space, (fupper-flower+7)/3);
         if (flast_d[d] != NULL)
             flast_d[d] = flast_d[d] - flower + 2;
            else
              return return_val;
         }
      } //while d<=MAX_D
    
  if (S!=NULL) { /*trace back*/
      int row1, col1, diag1, diag;
      d = return_val; diag = best_diag;
      row = *e2; col = *e1;
      while (d > 0) {
         diag1 = get_last(flast_d, d, diag, &row1);
         col1 = row1+diag1-ORIGIN;
         if (diag1 == diag) {
                if (row-row1 > 0) edit_script_rep(S, row-row1);
                } else if (diag1 < diag) {
                if (row-row1 > 0) edit_script_rep(S, row-row1);
                edit_script_ins(S,1);
                } else {
                if (row-row1-1> 0) edit_script_rep(S, row-row1-1);
                edit_script_del(S, 1);
                }
         d--; diag = diag1; col = col1; row = row1;
         } //while d>0
      edit_script_rep(S, flast_d[0][ORIGIN]);
      if (!reverse) 
            edit_script_reverse_inplace(S);
      }
    return return_val;
}


#define GAPALIGN_SUB ((unsigned char)0)  /*op types within the edit script*/
#define GAPALIGN_INS ((unsigned char)1)
#define GAPALIGN_DEL ((unsigned char)2)
#define GAPALIGN_DECLINE ((unsigned char)3)
char xgapcodes[4]={'S','I', 'D', 'X'};

struct GapXEditScript {
    unsigned char op_type;  // GAPALIGN_SUB, GAPALIGN_INS, or GAPALIGN_DEL
    int num;                // Number of operations
    GapXEditScript* next;
    GapXEditScript() {
        op_type=0;
        num=0;
        next=NULL;
        }
};

class CSeqGap { //
 public:
   int offset;
   int len;
   CSeqGap(int gofs=0,int glen=1) {
     offset=gofs;
     len=glen;
     }
};

class CAlnGapInfo {
    int a_ofs; //alignment start on seq a (0 based)
    int b_ofs; //alignment start on seq b (0 based)
    int a_len; //length of alignment on seq a
    int b_len; //length of alignment on seq b
  public:
    GVec<CSeqGap> a_gaps;
    GVec<CSeqGap> b_gaps;
    CAlnGapInfo(SEditScript* ed_script, int astart=0, int bstart=0):a_gaps(),b_gaps() {
       a_ofs=astart;
       b_ofs=bstart;
       a_len=0;
       b_len=0;
       for (int i=0; i<ed_script->num; i++) {
          int num=((ed_script->op[i]) >> 2);
          char op_type = 3 - ( ed_script->op[i] & EDIT_OP_MASK );
          if (op_type == 3 || op_type < 0 )
             GError("Error: encountered op_type %d in ed_script?!\n", (int)op_type);
          CSeqGap gap;
          switch (op_type) {
             case GAPALIGN_SUB: a_len+=num;
                                b_len+=num;
                                break;
             case GAPALIGN_INS: a_len+=num;
                                gap.offset=b_ofs+b_len;
                                gap.len=num;
                                b_gaps.Add(gap);
                                break;
             case GAPALIGN_DEL: b_len+=num;
                                gap.offset=a_ofs+a_len;
                                gap.len=num;
                                a_gaps.Add(gap);
                                break;
             }
          }
       }
    void printSeqA(const char* s) {
       int g=0;
       int aend=a_ofs+a_len;
       for (int i=a_ofs;i<aend;i++) {
         if (g<a_gaps.Count() && a_gaps[g].offset==i) {
            for (int j=0;j<a_gaps[g].len;j++) fprintf(stderr, "-");
            g++;
            }
         fprintf(stderr, "%c", s[i]);
         }
       fprintf(stderr, "\n");
       }
    void printSeqB(const char* s) {
       int g=0;
       int bend=b_ofs+b_len;
       for (int i=b_ofs;i<bend;i++) {
         if (g<b_gaps.Count() && b_gaps[g].offset==i) {
            for (int j=0;j<b_gaps[g].len;j++) fprintf(stderr, "-");
            g++;
            }
         fprintf(stderr, "%c", s[i]);
         }
       fprintf(stderr, "\n");
       }
  };
  
GapXEditScript* EditScript2GapX(SEditScript* ed_script) {
   GapXEditScript* esp_array=NULL;
   uint i;
   if (ed_script==NULL || ed_script->num == 0)
      return NULL;
   //GMALLOC(esp_array, ed_script->num * sizeof(GapXEditScript));
   esp_array = new GapXEditScript[ed_script->num];
   for (i=0; i<ed_script->num; i++) {
      //GCALLOC(esp, sizeof(GapXEditScript));
      //esp->num = EDIT_VAL(ed_script->op[i]); 
      //esp->num = ((ed_script->op[i]) >> 2);
      esp_array[i].num=((ed_script->op[i]) >> 2);
      //esp->op_type = 3 - EDIT_OPC(ed_script->op[i]);
      //esp->op_type = 3 - ( ed_script->op[i] & EDIT_OP_MASK );
      esp_array[i].op_type = 3 - ( ed_script->op[i] & EDIT_OP_MASK );
      if (esp_array[i].op_type == 3)
         //fprintf(stderr, "op_type = 3\n");
         GError("Error: EditScript2GapX encountered op_type 3 ?!\n");
      if (i>0)
         esp_array[i-1].next=&esp_array[i];
      }
   return esp_array;
}



struct GappedHSPInfo {
 const char *qseq;
 int ql,qr;
 const char *sseq;
 int sl,sr;
 int score;
 double pid;
 GapXEditScript* editscript;
 CAlnGapInfo* gapinfo;
 GappedHSPInfo(const char* q, int q_l, int q_r, const char* s, int s_l, int s_r,
      int sc=0, double percid=0) {
    qseq=q;
    sseq=s;
    ql=q_l;
    qr=q_r;
    sl=s_l;
    sr=s_r;
    score=sc;
    pid=percid;
    editscript=NULL;
    gapinfo=NULL;
    }
  ~GappedHSPInfo() {
    delete[] editscript;
    delete gapinfo;
    }
};

GappedHSPInfo* GreedyAlign(const char* q_seq,  int q_off, const char* s_seq, int s_off,
                    int reward, int penalty, int xdrop) {
  
  SEditScript* ed_script_fwd = edit_script_new();
  SEditScript* ed_script_rev = edit_script_new();
  int q_ext_l=0, q_ext_r=0, s_ext_l=0, s_ext_r=0;
  int score=0;
  int slen=strlen(s_seq); //subj
  int qlen=strlen(q_seq); //query
  const char* q=q_seq+q_off;
  int q_avail=qlen-q_off;
  const char* s=s_seq+s_off;
  int s_avail=slen-s_off;
  int MIN_GREEDY_SCORE=5; //minimum score for an alignment to be reported
  GreedyAlignMem* abmp=new GreedyAlignMem(slen, reward, -penalty, xdrop);
  // extend to the right
  score = MegaBlastGreedyAlign(s, s_avail, q, q_avail, false, xdrop,
             reward, -penalty, &s_ext_r, &q_ext_r, abmp, ed_script_fwd);
  GMessage("fwd score=%d (up to a_r=%d, b_r=%d)\n", score, q_ext_r, s_ext_r);
  /* extend to the left */
  score += MegaBlastGreedyAlign(s_seq, s_off, q_seq, q_off, true, xdrop,
              reward, -penalty, &s_ext_l, &q_ext_l, abmp, ed_script_rev);
  GMessage("rvs score=%d (back to a_l=%d, b_l=%d)\n", score, q_ext_l, s_ext_l);
  delete abmp;            
  GappedHSPInfo* ghsp=NULL;
  // In basic case the greedy algorithm returns number of
  // differences, hence we need to convert it to score     
  int retscore = (q_ext_r + s_ext_r + q_ext_l + s_ext_l)*reward/2 - score*(reward-penalty);
  GMessage("Actual score: %d\n",retscore);
  edit_script_append(ed_script_rev, ed_script_fwd); //combine the two extensions
  if (retscore>=MIN_GREEDY_SCORE) {
    ghsp=new GappedHSPInfo(q_seq, q_ext_l,q_ext_r, s_seq, s_ext_l, s_ext_r);
    int hsp_length = GMIN(q_ext_l+q_ext_r, s_ext_l+s_ext_r);
    ghsp->score=retscore;
    ghsp->pid = 100 * (1 - ((double) score) / hsp_length);
    //ghsp->editscript = EditScript2GapX(ed_script_rev);
    
    ghsp->gapinfo = new CAlnGapInfo(ed_script_rev, ghsp->ql, ghsp->sl);
    edit_script_free(ed_script_rev);
    edit_script_free(ed_script_fwd);
    }
  return ghsp;
 }

#define FRCLOSE(fh) if (fh!=NULL && fh!=stdin) fclose(fh)
#define FWCLOSE(fh) if (fh!=NULL && fh!=stdout) fclose(fh)

char* seq_cleanup(const char* s) {
 char* r=NULL;
 int slen=strlen(s);
 GMALLOC(r, slen+1);
 int newlen=0;
 for (int i=0;i<slen;i++) {
   if (s[i]<'A' || s[i]>'z' || (s[i]>'Z' && s[i]<'a')) continue;
   r[newlen]=s[i];
   newlen++;
   }
 r[newlen]=0;
 return r;
}

void loadSeqs(const char* fname, char*& s1, char*& s2) {
 GLineReader lr(fname);
 char* l=NULL;
 s1=NULL;
 s2=NULL;
 while ((l=lr.nextLine())!=NULL) {
  if (lr.length()<=3 || l[0]=='#') continue;
  if (s1==NULL) {
    s1=seq_cleanup(l);
    }
   else if (s2==NULL) s2=seq_cleanup(l);
           else break;
  }
}

//================= MAIN =====================

int main(int argc, char * const argv[]) {
 GArgs args(argc, argv, "a:b:i:");
 int e;
 if ((e=args.isError())>0)
    GError("%s\nInvalid argument: %s\n", USAGE, argv[e]);
 if (args.getOpt('h')!=NULL) GError("%s\n", USAGE);
 char* seqa=NULL;
 char* seqb=NULL;
 GStr s=args.getOpt('i');
 if (!s.is_empty()) loadSeqs(s.chars(), seqa, seqb);
   else {
     s=args.getOpt('a');
     if (!s.is_empty()) seqa=seq_cleanup(s.chars());     
     s=args.getOpt('b');
     if (!s.is_empty()) seqb=seq_cleanup(s.chars());     
     }
 if (seqa==NULL || seqb==NULL) GError("Error: both sequences are required!\n");
 GMessage("seqa: %s\n",seqa);
 GMessage("seqb: %s\n",seqb);
 //offsets must be in the middle of a good (best) HSP (ungapped match)
 GappedHSPInfo* ghsp=GreedyAlign(seqa, 0, seqb, 0, 2, -3, 22);
 if (ghsp!=NULL) {
   GMessage("a(%d..%d) align to b(%d..%d), score=%d, pid=%4.2f\n", ghsp->ql, ghsp->qr, 
                        ghsp->sl, ghsp->sr, ghsp->score, ghsp->pid);
   if (ghsp->editscript!=NULL) {
     GMessage("Edit script: ");
     GapXEditScript* ges=ghsp->editscript;
     do {
       GMessage("%d%c ",ges->num, xgapcodes[ges->op_type]); 
       } while ((ges=ges->next)!=NULL); 
     GMessage("\n");
     }
   if (ghsp->gapinfo!=NULL) {
     GMessage("Alignment:\n");
     ghsp->gapinfo->printSeqA(seqa);
     ghsp->gapinfo->printSeqB(seqb);
     }
   delete ghsp;
   }
 //---- done here
 GFREE(seqa);
 GFREE(seqb);
}


